import numpy as np
from buildingsysid.model_set.grey.base_greybox import BaseGreyBox
from buildingsysid.utils.statespace import StateSpace


# =================================================================
# FIRST ORDER MODEL
# =================================================================
class First(BaseGreyBox):
    
    def __init__(self, floor_area=1, fixed_params={}, param_bounds={}):
        super().__init__()
        
        self.floor_area = floor_area
        
        # Define total parameters and state dimensions
        self.total_parameters = 6
        self.n_states = 1
        
        # Define parameter dictionary
        self.param_dict = {
            0: ("H_1a", "W/(K*m²)", self.floor_area),           # Heat transfer coefficient per floor area
            1: ("C_1", "Wh/(K*m²)", self.floor_area * 3600),    # Specific heat capacity
            2: ("ws1", "%", self.floor_area / 100),             # window to floor percentage
            3: ("wh1", "-", 1.0),                               # heat input factor
            4: ("k1", "-", 1.0),                                # kalman gain
            5: ("x1[0]", "°C", 1.0)                             # initial state
        }
        
        self._initialize_fixed_parameters(fixed_params, param_bounds)
    
    
    def create_model(self, reduced_par):
        """
        Create the state-space model with the given parameters.
        
        Args:
            reduced_par: Array of free parameter values
            
        Returns:
            StateSpace: State-space model
        """      
        
        # Expand the reduced parameter vector to the full parameter vector
        full_model_par = self._expand_parameters(reduced_par)
        
        # Extract scaled parameters
        H_1a = full_model_par[0] * self.param_dict[0][2]    # (W/K)
        C_1 = full_model_par[1] * self.param_dict[1][2]     # (J/K)
        ws1 = full_model_par[2] * self.param_dict[2][2]     # (m²)
        wh1 = full_model_par[3] * self.param_dict[3][2]     # (-)
        k1 = full_model_par[4] * self.param_dict[4][2]      # (-)
        x1_0 = full_model_par[5] * self.param_dict[5][2]    # (°C)
        
        # Initial state vector
        x0 = np.array([[x1_0]])
        
        # Calculate the A matrix
        A = np.array([[-H_1a / C_1]])
        
        # Construct the B matrix (inputs: ambient temp, solar radiation, heating)
        B = np.array([[H_1a / C_1, ws1 / C_1, wh1 / C_1]])
        
        C = np.array([[1]])
        
        D = np.array([[0, 0, 0]])
        
        K = np.array([[k1]])
            
        return StateSpace(A, B, C, D, K, x0, samplingTime=0)
    
    
# =================================================================
# Second-order model
# =================================================================
class Second(BaseGreyBox):
    
    def __init__(self, floor_area=1, fixed_params={}, param_bounds={}):
        super().__init__()
        
        self.floor_area = floor_area
        
        # Define states and maximum parameters
        self.total_parameters = 13
        self.n_states = 2
           
        # Define parameter dictionary
        self.param_dict = {
            0: ("H_12", "W/(K*m²)", self.floor_area),
            1: ("H_1a", "W/(K*m²)", self.floor_area),
            2: ("H_2a", "W/(K*m²)", self.floor_area),
            3: ("C_1", "Wh/(K*m²)", self.floor_area * 3600),
            4: ("C_2", "Wh/(K*m²)", self.floor_area * 3600),
            5: ("ws1", "%", self.floor_area / 100),
            6: ("ws2", "%", self.floor_area / 100),
            7: ("wh1", "-", 1.0),
            8: ("wh2", "-", 1.0),
            9: ("k1", "-", 1.0),
            10: ("k2", "-", 1.0),
            11: ("x1[0]", "°C", 1.0),
            12: ("x2[0]", "°C", 1.0)
        }
        
        self._initialize_fixed_parameters(fixed_params, param_bounds)
    
    
    def create_model(self, reduced_par):
        """Create the second-order state-space model"""
        
        # Expand the reduced parameter vector to the full parameter vector
        full_model_par = self._expand_parameters(reduced_par)
        
        # Extract scaled parameters
        H_12 = full_model_par[0] * self.param_dict[0][2]    # (W/K)
        H_1a = full_model_par[1] * self.param_dict[1][2]    # (W/K)
        H_2a = full_model_par[2] * self.param_dict[2][2]    # (W/K)
        C_1 = full_model_par[3] * self.param_dict[3][2]     # (J/K)
        C_2 = full_model_par[4] * self.param_dict[4][2]     # (J/K)
        ws1 = full_model_par[5] * self.param_dict[5][2]     # (m²)
        ws2 = full_model_par[6] * self.param_dict[6][2]     # (m²)
        wh1 = full_model_par[7] * self.param_dict[7][2]     # (-)
        wh2 = full_model_par[8] * self.param_dict[8][2]     # (-)
        k1 = full_model_par[9] * self.param_dict[9][2]      # (-)
        k2 = full_model_par[10] * self.param_dict[10][2]    # (-)
        x1_0 = full_model_par[11] * self.param_dict[11][2]  # (°C)
        x2_0 = full_model_par[12] * self.param_dict[12][2]  # (°C)
        
        
        # Calculate the elements of the A matrix
        a11 = -(H_1a + H_12) / C_1
        a12 = H_12 / C_1
        a21 = H_12 / C_2
        a22 = -(H_2a + H_12) / C_2
        
        # Construct the matrix A
        A = np.array([
            [a11, a12],
            [a21, a22]
        ])
        
        # Construct the B matrix (inputs: ambient temp, solar radiation, heating)
        B = np.array([
            [H_1a / C_1, ws1 / C_1, wh1 / C_1],
            [H_2a / C_2, ws2 / C_2, wh2 / C_2]
        ])
        
        C = np.array([[1, 0]])
        
        D = np.array([[0, 0, 0]])
        
        x0 = np.array([[x1_0],
                       [x2_0]])
        
        K = np.array([[k1],
                      [k2]])
            
        return StateSpace(A, B, C, D, K, x0, samplingTime=0)
    
    
# =================================================================
# Third-order model
# =================================================================    
class Third(BaseGreyBox):
    
    def __init__(self, floor_area=1, fixed_params={}, param_bounds={}):
        super().__init__()
        
        self.floor_area = floor_area
        
        # Define states and maximum parameters
        self.n_states = 3
        self.total_parameters = 21
        
        # Define parameter dictionary
        self.param_dict = {
            0: ("H_12", "W/(K*m²)", self.floor_area),
            1: ("H_13", "W/(K*m²)", self.floor_area),
            2: ("H_23", "W/(K*m²)", self.floor_area),
            3: ("H_1a", "W/(K*m²)", self.floor_area),
            4: ("H_2a", "W/(K*m²)", self.floor_area),
            5: ("H_3a", "W/(K*m²)", self.floor_area),
            6: ("C_1", "Wh/(K*m²)", self.floor_area * 3600),
            7: ("C_2", "Wh/(K*m²)", self.floor_area * 3600),
            8: ("C_3", "Wh/(K*m²)", self.floor_area * 3600),
            9: ("ws1", "%", self.floor_area / 100),
            10: ("ws2", "%", self.floor_area / 100),
            11: ("ws3", "%", self.floor_area / 100),
            12: ("wh1", "-", 1.0),
            13: ("wh2", "-", 1.0),
            14: ("wh3", "-", 1.0),
            15: ("k1", "-", 1.0),
            16: ("k2", "-", 1.0),
            17: ("k3", "-", 1.0),
            18: ("x1[0]", "°C", 1.0),
            19: ("x2[0]", "°C", 1.0),
            20: ("x3[0]", "°C", 1.0)
        }
        
        self._initialize_fixed_parameters(fixed_params, param_bounds)
        
    
    def create_model(self, reduced_par):
        """Create the second-order state-space model"""
        
        # Expand parameters if needed
        full_model_par = self._expand_parameters(reduced_par)
        
        # Use the scales defined in param_dict
        H_12 = full_model_par[0] * self.param_dict[0][2]  # (W/K)
        H_13 = full_model_par[1] * self.param_dict[1][2]  # (W/K)
        H_23 = full_model_par[2] * self.param_dict[2][2]  # (W/K)
        H_1a = full_model_par[3] * self.param_dict[3][2]  # (W/K)
        H_2a = full_model_par[4] * self.param_dict[4][2]  # (W/K)
        H_3a = full_model_par[5] * self.param_dict[5][2]  # (W/K)
        C_1 = full_model_par[6] * self.param_dict[6][2]   # (J/K)
        C_2 = full_model_par[7] * self.param_dict[7][2]   # (J/K)
        C_3 = full_model_par[8] * self.param_dict[8][2]   # (J/K)
        ws1 = full_model_par[9] * self.param_dict[9][2]   # (m²)
        ws2 = full_model_par[10] * self.param_dict[10][2]   # (m²) 
        ws3 = full_model_par[11] * self.param_dict[11][2]   # (m²) 
        wh1 = full_model_par[12] * self.param_dict[12][2]   # (-)
        wh2 = full_model_par[13] * self.param_dict[13][2]   # (-)
        wh3 = full_model_par[14] * self.param_dict[14][2]   # (-)
        k1 = full_model_par[15] * self.param_dict[15][2]   # (-)
        k2 = full_model_par[16] * self.param_dict[16][2]   # (-)
        k3 = full_model_par[17] * self.param_dict[17][2]   # (-)
        x1_0 = full_model_par[18] * self.param_dict[18][2]   # (°C)
        x2_0 = full_model_par[19] * self.param_dict[19][2]   # (°C)
        x3_0 = full_model_par[20] * self.param_dict[20][2]   # (°C)
        
        # Initial state
        x0 = np.array([[x1_0],
                      [x2_0],
                      [x3_0]])
        
        # Calculate the elements of the matrix
        a11 = -(H_1a + H_12 + H_13) / C_1
        a12 = H_12 / C_1
        a13 = H_13 / C_1
        
        a21 = H_12 / C_2
        a22 = -(H_2a + H_12 + H_23) / C_2
        a23 = H_23 / C_2
        
        a31 = H_13 / C_3
        a32 = H_23 / C_3
        a33 = -(H_3a + H_13 + H_23) / C_3
        
        A = np.array([
            [a11, a12, a13],
            [a21, a22, a23],
            [a31, a32, a33]
        ])
                
        B = np.array([
            [H_1a / C_1, ws1 / C_1, wh1 / C_1],
            [H_2a / C_2, ws2 / C_2, wh2 / C_2],
            [H_3a / C_3, ws3 / C_3, wh3 / C_3]
        ])
        
        C = np.array([[1, 0, 0]])
        D = np.array([[0, 0, 0]])
        
        K = np.array([[k1],
                      [k2],
                      [k3]])
            
        return StateSpace(A, B, C, D, K, x0, samplingTime=0)
    
    
# =================================================================
# Fourth-order model
# =================================================================    
class Fourth(BaseGreyBox):
    
    def __init__(self, floor_area=1, fixed_params={}, param_bounds={}):
        super().__init__()
        
        self.floor_area = floor_area
        
        # Define states and maximum parameters
        self.n_states = 4
        self.total_parameters = 30
        
        # Define parameter dictionary for printing: (name, unit, scale_factor)
        self.param_dict = {
            0: ("H_12", "W/(K*m²)", self.floor_area),
            1: ("H_13", "W/(K*m²)", self.floor_area),
            2: ("H_14", "W/(K*m²)", self.floor_area),
            3: ("H_23", "W/(K*m²)", self.floor_area),
            4: ("H_24", "W/(K*m²)", self.floor_area),
            5: ("H_34", "W/(K*m²)", self.floor_area),
            6: ("H_1a", "W/(K*m²)", self.floor_area),
            7: ("H_2a", "W/(K*m²)", self.floor_area),
            8: ("H_3a", "W/(K*m²)", self.floor_area),
            9: ("H_4a", "W/(K*m²)", self.floor_area),
            10: ("C_1", "Wh/(K*m²)", self.floor_area * 3600),
            11: ("C_2", "Wh/(K*m²)", self.floor_area * 3600),
            12: ("C_3", "Wh/(K*m²)", self.floor_area * 3600),
            13: ("C_4", "Wh/(K*m²)", self.floor_area * 3600),
            14: ("ws1", "%", self.floor_area / 100),
            15: ("ws2", "%", self.floor_area / 100),
            16: ("ws3", "%", self.floor_area / 100),
            17: ("ws4", "%", self.floor_area / 100),
            18: ("wh1", "-", 1.0),
            19: ("wh2", "-", 1.0),
            20: ("wh3", "-", 1.0),
            21: ("wh4", "-", 1.0),
            22: ("k1", "-", 1.0),
            23: ("k2", "-", 1.0),
            24: ("k3", "-", 1.0),
            25: ("k4", "-", 1.0),
            26: ("x1[0]", "°C", 1.0),
            27: ("x2[0]", "°C", 1.0),
            28: ("x3[0]", "°C", 1.0),
            29: ("x4[0]", "°C", 1.0)
        }
    
    def create_model(self, reduced_par):
        """Create the fourth-order state-space model"""
        
        # Expand parameters if needed
        full_model_par = self._expand_parameters(reduced_par)
        
        # Extract scaled parameters
        H_12 = full_model_par[0] * self.param_dict[0][2]  # (W/K)
        H_13 = full_model_par[1] * self.param_dict[1][2]  # (W/K)
        H_14 = full_model_par[2] * self.param_dict[2][2]  # (W/K)
        H_23 = full_model_par[3] * self.param_dict[3][2]  # (W/K)
        H_24 = full_model_par[4] * self.param_dict[4][2]  # (W/K)
        H_34 = full_model_par[5] * self.param_dict[5][2]  # (W/K)
        H_1a = full_model_par[6] * self.param_dict[6][2]  # (W/K)
        H_2a = full_model_par[7] * self.param_dict[7][2]  # (W/K)
        H_3a = full_model_par[8] * self.param_dict[8][2]  # (W/K)
        H_4a = full_model_par[9] * self.param_dict[9][2]  # (W/K)
        C_1 = full_model_par[10] * self.param_dict[10][2]  # (J/K)
        C_2 = full_model_par[11] * self.param_dict[11][2]  # (J/K)
        C_3 = full_model_par[12] * self.param_dict[12][2]  # (J/K)
        C_4 = full_model_par[13] * self.param_dict[13][2]  # (J/K)
        ws1 = full_model_par[14] * self.param_dict[14][2]  # (m²)
        ws2 = full_model_par[15] * self.param_dict[15][2]  # (m²)
        ws3 = full_model_par[16] * self.param_dict[16][2]  # (m²)
        ws4 = full_model_par[17] * self.param_dict[17][2]  # (m²)
        wh1 = full_model_par[18] * self.param_dict[18][2]  # (-)
        wh2 = full_model_par[19] * self.param_dict[19][2]  # (-)
        wh3 = full_model_par[20] * self.param_dict[20][2]  # (-)
        wh4 = full_model_par[21] * self.param_dict[21][2]  # (-)
        k1 = full_model_par[22] * self.param_dict[22][2]  # (-)
        k2 = full_model_par[23] * self.param_dict[23][2]  # (-)
        k3 = full_model_par[24] * self.param_dict[24][2]  # (-)
        k4 = full_model_par[25] * self.param_dict[25][2]  # (-)
        x1_0 = full_model_par[26] * self.param_dict[26][2]  # (°C)
        x2_0 = full_model_par[27] * self.param_dict[27][2]  # (°C)
        x3_0 = full_model_par[28] * self.param_dict[28][2]  # (°C)
        x4_0 = full_model_par[29] * self.param_dict[29][2]  # (°C)
        
        x0 = np.array([
            [x1_0],
            [x2_0],        
            [x3_0],        
            [x4_0]
        ])
        
        a11 = -(H_1a + H_12 + H_13 + H_14) / C_1
        a12 = H_12 / C_1
        a13 = H_13 / C_1
        a14 = H_14 / C_1
        
        a21 = H_12 / C_2
        a22 = -(H_2a + H_12 + H_23 + H_24) / C_2
        a23 = H_23 / C_2
        a24 = H_24 / C_2
        
        a31 = H_13 / C_3
        a32 = H_23 / C_3
        a33 = -(H_3a + H_13 + H_23 + H_34) / C_3
        a34 = H_34 / C_3
        
        a41 = H_14 / C_4
        a42 = H_24 / C_4
        a43 = H_34 / C_4
        a44 = -(H_4a + H_14 + H_24 + H_34) / C_4
        
        A = np.array([
            [a11, a12, a13, a14],
            [a21, a22, a23, a24],
            [a31, a32, a33, a34],
            [a41, a42, a43, a44]
        ])
        
        B = np.array([
            [H_1a / C_1, ws1 / C_1, wh1 / C_1],
            [H_2a / C_2, ws2 / C_2, wh2 / C_2],
            [H_3a / C_3, ws3 / C_3, wh3 / C_3],
            [H_4a / C_4, ws4 / C_4, wh4 / C_4]
        ])
        
        C = np.array([[1, 0, 0, 0]])
        
        D = np.array([[0, 0, 0]])
        
        K = np.array([
            [k1],
            [k2],        
            [k3],        
            [k4]
        ])
            
        return StateSpace(A, B, C, D, K, x0, samplingTime=0)


from buildingsysid.data.iddata_base import IDData

from buildingsysid.data.iddata_resample import IDDataResample
from buildingsysid.data.iddata_split import IDDataSplit
from buildingsysid.data.iddata_timeseriesplot import IDDataTimeSeriesPlot
from buildingsysid.data.iddata_correlation import IDDataCorrelation
from buildingsysid.data.iddata_normalize import IDDataNormalize

class IDData(IDData, IDDataResample, IDDataSplit, IDDataTimeSeriesPlot, IDDataCorrelation, IDDataNormalize):
    """
    Initialize IDData object for system identification data.
    
    Args:
        y (numpy.ndarray): Output data with shape (num_outputs, num_samples) or (num_samples,)
        u (numpy.ndarray): Input data with shape (num_inputs, num_samples) or (num_samples,)
        samplingTime (float): Sampling time in seconds
        timestamps (array-like): Timestamps for the data points
        y_names (list, optional): Names for each output channel
        u_names (list, optional): Names for each input channel
        y_units (list, optional): Units for each output channel
        u_units (list, optional): Units for each input channel
    """
    pass


import numpy as np
from scipy.optimize import least_squares, differential_evolution, minimize


# =====================================================================
# Define solvers - these would typically be in a separate module
# =====================================================================

def least_squares_solver(objective_fn, x0, bounds, options=None):
    """
    Solver using scipy's least_squares optimizer.
    
    Args:
        objective_fn: Function that calculates residuals
        x0: Initial parameter values
        bounds: Tuple of (lower_bounds, upper_bounds)
        options: Dictionary of solver options
        
    Returns:
        Optimization result object
    """
    options = options or {}
    method = options.pop('method', 'trf')
    verbose = options.pop('verbose', 1)
    
    # Run optimization
    result = least_squares(
        objective_fn,
        x0,
        bounds=bounds,
        method=method,
        verbose=verbose,
        **options
    )
    
    return result


def global_optimizer_solver(objective_fn, x0, bounds, options=None):
    """
    Solver using differential evolution for global optimization.
    
    Args:
        objective_fn: Function that calculates residuals
        x0: Initial parameter values (not used by differential evolution)
        bounds: Tuple of (lower_bounds, upper_bounds)
        options: Dictionary of solver options
        
    Returns:
        Optimization result object
    """
    options = options or {}
    popsize = options.pop('popsize', 15)
    maxiter = options.pop('maxiter', 100)
    disp = options.pop('disp', True)
    
    # Prepare bounds for differential_evolution
    de_bounds = list(zip(bounds[0], bounds[1]))
    
    # Define sum of squares objective
    def sum_sq_objective(x):
        residuals = objective_fn(x)
        return np.sum(residuals**2)
    
    # Run optimization
    result = differential_evolution(
        sum_sq_objective,
        de_bounds,
        popsize=popsize,
        maxiter=maxiter,
        disp=disp,
        **options
    )
    
    return result


def custom_solver(objective_fn, x0, bounds, options=None):
    """
    Custom solver implementation using a simple gradient-free approach.
    This is just a simple example - a real implementation would be more sophisticated.
    
    Args:
        objective_fn: Function that calculates residuals
        x0: Initial parameter values
        bounds: Tuple of (lower_bounds, upper_bounds)
        options: Dictionary of solver options
        
    Returns:
        Object with optimization results
    """
    options = options or {}
    max_iter = options.get('max_iter', 100)
    step_size = options.get('step_size', 0.1)
    tolerance = options.get('tolerance', 1e-6)
    
    # Initialize best solution
    x_best = np.array(x0, dtype=float)
    f_best = np.sum(objective_fn(x_best)**2)  # Sum of squares
    
    # Track progress
    iterations = 0
    converged = False
    
    # Simple coordinate descent algorithm
    while iterations < max_iter and not converged:
        iterations += 1
        x_prev = x_best.copy()
        
        # Try to improve each parameter
        for i in range(len(x_best)):
            # Try increasing
            x_try = x_best.copy()
            x_try[i] += step_size
            # Ensure within bounds
            x_try[i] = min(x_try[i], bounds[1][i])
            
            # Evaluate
            f_try = np.sum(objective_fn(x_try)**2)
            if f_try < f_best:
                x_best = x_try
                f_best = f_try
                continue
                
            # Try decreasing
            x_try = x_best.copy()
            x_try[i] -= step_size
            # Ensure within bounds
            x_try[i] = max(x_try[i], bounds[0][i])
            
            # Evaluate
            f_try = np.sum(objective_fn(x_try)**2)
            if f_try < f_best:
                x_best = x_try
                f_best = f_try
        
        # Check convergence
        change = np.linalg.norm(x_best - x_prev)
        if change < tolerance:
            converged = True
            
        # Print progress
        if iterations % 10 == 0 or converged:
            print(f"Iteration {iterations}: Best SSE = {f_best:.6f}, Change = {change:.6f}")
    
    # Create a result object similar to scipy's
    class OptimResult:
        pass
    
    result = OptimResult()
    result.x = x_best
    result.fun = objective_fn(x_best)
    result.success = converged
    result.message = "Converged" if converged else "Maximum iterations reached"
    result.nit = iterations
    
    return result





class IDDataSplit:
    """Mixin with resampling functionality"""

    def split(self, train_ratio=0.7):
        """
        Split the time series data into training and validation sets.
        
        For time series data, this maintains the temporal order by using
        the earlier portion for training and the later portion for validation.
        
        Args:
            train_ratio (float): Proportion of data to use for training (0 to 1)
            
        Returns:
            tuple: (train_data, val_data) - Two IDData objects containing the split data
        """
        # Determine split index
        n_samples = len(self.timestamps)
        split_idx = int(n_samples * train_ratio)
        
        # Split the data - arrays are guaranteed to be 2D at this point
        y_train = self.y[:, :split_idx]
        y_val = self.y[:, split_idx:]
        u_train = self.u[:, :split_idx]
        u_val = self.u[:, split_idx:]

        
        # Explicitly pass ALL metadata to the new objects
        train_data = self.__class__(
            y=y_train,
            u=u_train,
            samplingTime=self.samplingTime,
            timestamps=self.timestamps[:split_idx],
            y_names=list(self.y_names),  # Create new copies using list()
            u_names=list(self.u_names),
            y_units=list(self.y_units),
            u_units=list(self.u_units)
        )
        
        val_data = self.__class__(
            y=y_val,
            u=u_val,
            samplingTime=self.samplingTime,
            timestamps=self.timestamps[split_idx:],
            y_names=list(self.y_names),
            u_names=list(self.u_names),
            y_units=list(self.y_units),
            u_units=list(self.u_units)
        )
        
        # Transfer normalization information if present
        if self.is_normalized:
            train_data.y_norm_params = self.y_norm_params
            train_data.u_norm_params = self.u_norm_params
            train_data.is_normalized = True
            
            val_data.y_norm_params = self.y_norm_params
            val_data.u_norm_params = self.u_norm_params
            val_data.is_normalized = True
        
        return train_data, val_data


from setuptools import setup, find_packages

if __name__ == "__main__":
    setup(
        packages=find_packages(include=['buildingsysid', 'buildingsysid.*'])
        )

# from setuptools import setup, find_packages

# setup(
#     name="buildingsysid",
#     version="0.1.1",
#     description="System identification tools for building models",
#     long_description=open("README.md").read(),
#     long_description_content_type="text/markdown",
#     author="Michael Dahl Knudsen",
#     author_email="mdk@cae.au.dk",
#     url="https://github.com/michael666nan/buildingsysid",
#     packages=find_packages(),
#     classifiers=[
#         "Programming Language :: Python :: 3",
#         "License :: OSI Approved :: MIT License",
#         "Operating System :: OS Independent",
#     ],
#     python_requires=">=3.7",
#     install_requires=[
#         "numpy",
#         "pandas",
#         "matplotlib",
#         "scipy",
#         "statsmodels",
#     ],
# )


import datetime
import numpy as np
import copy
import pandas as pd
from scipy import interpolate

class IDDataResample:
    """Mixin with resampling functionality"""

    def resample(self, new_sampling_time=None, method='linear', aggregation=None, 
             output_agg=None, input_agg=None):
        """
        Resample the input and output data to a new sampling time.
        
        Args:
            new_sampling_time (float, optional): New sampling time in seconds. 
                                                If None, returns a copy of the original object.
            method (str, optional): Interpolation method. Options:
                                   'linear' - Linear interpolation (default)
                                   'nearest' - Nearest neighbor interpolation
                                   'cubic' - Cubic spline interpolation
            aggregation (str, optional): Default aggregation method for downsampling. Options:
                                        None - Use interpolation for both up/downsampling (default)
                                        'first' - Use first value in each interval
                                        'last' - Use last value in each interval
                                        'mean' - Use mean value in each interval
                                        'max' - Use maximum value in each interval
                                        'min' - Use minimum value in each interval
            output_agg (str or list, optional): Specific aggregation method(s) for output channels.
                                               Can be a single method or a list with one method per channel.
            input_agg (str or list, optional): Specific aggregation method(s) for input channels.
                                              Can be a single method or a list with one method per channel.
                                   
        Returns:
            IDData: A new IDData object with resampled data
        """
        
        # If no resampling needed, return a copy
        if new_sampling_time is None or new_sampling_time == self.samplingTime:
            return copy.deepcopy(self)
        
        # Determine if this is downsampling (new_sampling_time > old_sampling_time)
        is_downsampling = new_sampling_time > self.samplingTime
        
        # For downsampling with aggregation
        if is_downsampling and (aggregation or output_agg or input_agg):
            return self._downsample_with_aggregation(new_sampling_time, aggregation, output_agg, input_agg)
    
        
        # Get numeric versions of original timestamps
        orig_numeric_ts = self._convert_timestamps_to_numeric(self.timestamps)
        
        # Calculate the new timestamps
        start_time = self.timestamps[0]
        end_time = self.timestamps[-1]
        
        # Create new timestamps array based on the new sampling time
        if isinstance(start_time, pd.Timestamp):
            # For pandas Timestamp
            total_duration = (end_time - start_time).total_seconds()
            num_new_samples = int(total_duration / new_sampling_time) + 1
            new_timestamps = pd.date_range(
                start=start_time,
                periods=num_new_samples,
                freq=pd.Timedelta(seconds=new_sampling_time)
            )
        elif isinstance(start_time, datetime.datetime):
            # For Python datetime
            total_duration = (end_time - start_time).total_seconds()
            num_new_samples = int(total_duration / new_sampling_time) + 1
            new_timestamps = [start_time + datetime.timedelta(seconds=i*new_sampling_time) 
                            for i in range(num_new_samples)]
        elif np.issubdtype(type(start_time), np.datetime64):
            # For numpy datetime64
            total_duration = (end_time - start_time) / np.timedelta64(1, 's')
            num_new_samples = int(total_duration / new_sampling_time) + 1
            time_delta = np.timedelta64(int(new_sampling_time), 's')
            new_timestamps = np.array([start_time + i * time_delta for i in range(num_new_samples)])
        else:
            # For numeric timestamps
            total_duration = end_time - start_time
            num_new_samples = int(total_duration / new_sampling_time) + 1
            new_timestamps = np.linspace(start_time, end_time, num_new_samples)
        
        # Get numeric versions of new timestamps
        new_numeric_ts = self._convert_timestamps_to_numeric(new_timestamps)
        
        # Prepare empty arrays for resampled data
        new_y = np.zeros((self.n_outputs, len(new_timestamps)))
        new_u = np.zeros((self.n_inputs, len(new_timestamps)))
        
        # Resample each output channel
        for i in range(self.n_outputs):
            interpolator = interpolate.interp1d(
                orig_numeric_ts, self.y[i], 
                kind=method, bounds_error=False, fill_value='extrapolate'
            )
            new_y[i] = interpolator(new_numeric_ts)
        
        # Resample each input channel
        for i in range(self.n_inputs):
            interpolator = interpolate.interp1d(
                orig_numeric_ts, self.u[i], 
                kind=method, bounds_error=False, fill_value='extrapolate'
            )
            new_u[i] = interpolator(new_numeric_ts)
        
        # Create a new IDData object with the resampled data
        return self.__class__(
            y=new_y, 
            u=new_u, 
            samplingTime=new_sampling_time,
            timestamps=new_timestamps,
            y_names=self.y_names.copy(),
            u_names=self.u_names.copy(),
            y_units=self.y_units.copy(),
            u_units=self.u_units.copy()
        )

    def _convert_timestamps_to_numeric(self, timestamps):
        """Convert any timestamp format to numeric values for interpolation."""
        if isinstance(timestamps[0], pd.Timestamp):
            # Convert pandas Timestamp to seconds since first timestamp
            first_ts = timestamps[0]
            return np.array([(ts - first_ts).total_seconds() for ts in timestamps])
        elif isinstance(timestamps[0], datetime.datetime):
            # Convert datetime to seconds since first timestamp
            first_ts = timestamps[0]
            return np.array([(ts - first_ts).total_seconds() for ts in timestamps])
        elif np.issubdtype(type(timestamps[0]), np.datetime64):
            # Convert numpy datetime64 to seconds since first timestamp
            first_ts = timestamps[0]
            return np.array([(ts - first_ts) / np.timedelta64(1, 's') for ts in timestamps])
        else:
            # Already numeric
            return np.array(timestamps)

    def _downsample_with_aggregation(self, new_sampling_time, aggregation, output_agg=None, input_agg=None):
        """
        Downsample data using specified aggregation methods.
        
        Args:
            new_sampling_time (float): New sampling time in seconds
            aggregation (str): Default aggregation method for all channels
            output_agg (str or list): Aggregation method(s) for output channels
                                     Can be a single string or a list/tuple with one method per channel
            input_agg (str or list): Aggregation method(s) for input channels
                                    Can be a single string or a list/tuple with one method per channel
            
        Returns:
            IDData: Downsampled data object
        """
        # Set default aggregation method
        default_agg = aggregation or 'mean'
        
        # Process output aggregation methods
        if output_agg is None:
            # Use default for all output channels
            output_agg_methods = [default_agg] * self.n_outputs
        elif isinstance(output_agg, str):
            # Use the same specified method for all output channels
            output_agg_methods = [output_agg] * self.n_outputs
        else:
            # Use provided list of methods
            if len(output_agg) != self.n_outputs:
                print(f"Warning: Length of output_agg ({len(output_agg)}) doesn't match number of outputs ({self.n_outputs}). Using default.")
                output_agg_methods = [default_agg] * self.n_outputs
            else:
                output_agg_methods = output_agg
        
        # Process input aggregation methods
        if input_agg is None:
            # Use default for all input channels
            input_agg_methods = [default_agg] * self.n_inputs
        elif isinstance(input_agg, str):
            # Use the same specified method for all input channels
            input_agg_methods = [input_agg] * self.n_inputs
        else:
            # Use provided list of methods
            if len(input_agg) != self.n_inputs:
                print(f"Warning: Length of input_agg ({len(input_agg)}) doesn't match number of inputs ({self.n_inputs}). Using default.")
                input_agg_methods = [default_agg] * self.n_inputs
            else:
                input_agg_methods = input_agg
        
        # Convert data to DataFrame for easier handling
        output_dfs = []
        for i in range(self.n_outputs):
            df = pd.DataFrame({
                'timestamp': self.timestamps,
                f'y{i}': self.y[i]
            })
            output_dfs.append(df)
        
        input_dfs = []
        for i in range(self.n_inputs):
            df = pd.DataFrame({
                'timestamp': self.timestamps,
                f'u{i}': self.u[i]
            })
            input_dfs.append(df)
        
        # Create resampling rule based on timestamps type
        if isinstance(self.timestamps[0], pd.Timestamp):
            # For pandas timestamps, timestamps are already datetime
            rule = pd.Timedelta(seconds=new_sampling_time)
        elif isinstance(self.timestamps[0], datetime.datetime):
            # Convert Python datetime to pandas datetime
            for df in output_dfs + input_dfs:
                df['timestamp'] = pd.to_datetime(df['timestamp'])
            rule = pd.Timedelta(seconds=new_sampling_time)
        elif np.issubdtype(type(self.timestamps[0]), np.datetime64):
            # For numpy datetime64, pandas can handle it directly
            rule = pd.Timedelta(seconds=new_sampling_time)
        else:
            # For numeric timestamps, convert to pandas datetime
            # Assuming numeric timestamps are in seconds from some reference point
            for df in output_dfs + input_dfs:
                # Create an arbitrary date and add timestamp as seconds
                df['timestamp'] = pd.to_datetime('2000-01-01') + pd.to_timedelta(df['timestamp'], unit='s')
            rule = pd.Timedelta(seconds=new_sampling_time)
        
        # Convert aggregation method strings to actual functions
        valid_methods = {'first', 'last', 'mean', 'max', 'min'}
        
        # Resample each output dataframe with its specific aggregation method
        resampled_output_dfs = []
        for i, df in enumerate(output_dfs):
            agg_method = output_agg_methods[i]
            if agg_method not in valid_methods:
                print(f"Warning: Invalid aggregation method '{agg_method}' for output {i}. Using 'mean'.")
                agg_method = 'mean'
                
            df.set_index('timestamp', inplace=True)
            resampled_df = df.resample(rule).agg(agg_method)
            resampled_output_dfs.append(resampled_df)
        
        # Resample each input dataframe with its specific aggregation method
        resampled_input_dfs = []
        for i, df in enumerate(input_dfs):
            agg_method = input_agg_methods[i]
            if agg_method not in valid_methods:
                print(f"Warning: Invalid aggregation method '{agg_method}' for input {i}. Using 'mean'.")
                agg_method = 'mean'
                
            df.set_index('timestamp', inplace=True)
            resampled_df = df.resample(rule).agg(agg_method)
            resampled_input_dfs.append(resampled_df)
        
        # Get new timestamps (all dataframes should have same timestamps after resampling)
        new_timestamps = resampled_output_dfs[0].index
        
        # Convert timestamps back to original format if needed
        if not isinstance(self.timestamps[0], pd.Timestamp):
            if isinstance(self.timestamps[0], datetime.datetime):
                new_timestamps = new_timestamps.to_pydatetime()
            elif np.issubdtype(type(self.timestamps[0]), np.datetime64):
                new_timestamps = new_timestamps.to_numpy()
            else:
                # If original was numeric, convert back to seconds since reference
                new_timestamps = (new_timestamps - pd.Timestamp('2000-01-01')).total_seconds()
        
        # Extract resampled data
        new_y = np.zeros((self.n_outputs, len(new_timestamps)))
        for i, df in enumerate(resampled_output_dfs):
            new_y[i] = df[f'y{i}'].to_numpy()
        
        new_u = np.zeros((self.n_inputs, len(new_timestamps)))
        for i, df in enumerate(resampled_input_dfs):
            new_u[i] = df[f'u{i}'].to_numpy()
        
        # Create a new IDData object with the resampled data
        return self.__class__(
            y=new_y, 
            u=new_u, 
            samplingTime=new_sampling_time,
            timestamps=new_timestamps,
            y_names=self.y_names.copy(),
            u_names=self.u_names.copy(),
            y_units=self.y_units.copy(),
            u_units=self.u_units.copy()
        )


import numpy as np
import matplotlib.pyplot as plt

def discrete_step_response(ss, input_channels=None, N=72, plot=True, output_names=None, title=None):
    """
    Calculate and optionally plot the step response of a state-space model.
    
    Parameters:
    -----------
    ss : object
        Discrete State-space model object containing A, B, C, D matrices
    input_channels : list or None, optional
        List of input channels to use for the step response (default: None, meaning all inputs)
    N : int, optional
        Number of time steps to simulate (default: 72)
    plot : bool, optional
        Whether to plot the step response (default: True)
    output_names : list, optional
        Names of the output channels for the plot legend (default: None)
    title : str, optional
        Custom title for the plot (default: None)
        
    Returns:
    --------
    y_sim : dict
        Dictionary of simulated outputs for each input channel
    x_sim : dict
        Dictionary of state trajectories for each input channel
    fig : matplotlib.figure.Figure or None
        Figure object if plot=True, otherwise None
    ss_gains : dict
        Dictionary of steady-state gains for each input channel
    """
    
    A = ss.A 
    B = ss.B 
    C = ss.C 
    D = ss.D 
    
    # Determine which input channels to use
    if input_channels is None:
        input_channels = list(range(B.shape[1]))
    elif isinstance(input_channels, int):
        input_channels = [input_channels]
    
    # Create dictionaries to store results for each input channel
    y_sim = {}
    x_sim = {}
    ss_gains = {}
    
    # Calculate steady-state gains
    try:
        # Computing (I-A)^(-1) for steady-state calculation
        I = np.eye(A.shape[0])
        I_minus_A_inv = np.linalg.solve(I - A, I)  # More stable than direct inverse
    except np.linalg.LinAlgError:
        print("Warning: Matrix (I-A) is singular. Steady-state gains may not be accurate.")
        I_minus_A_inv = None
    
    # Simulate step response for each input channel
    for input_channel in input_channels:
        # Initialize
        y = np.zeros((C.shape[0], N))
        x = np.zeros((A.shape[0], N))
        y[:,0] = C @ x[:,0] + D[:,input_channel] * 0  # No input at t=0
        
        # Simulate system dynamics
        for k in range(N-1):        
            x[:,k+1] = A @ x[:,k] + B[:,input_channel]
            y[:,k+1] = C @ x[:,k+1] + D[:,input_channel]
        
        # Store results
        y_sim[input_channel] = y
        x_sim[input_channel] = x
        
        # Calculate steady-state gains
        if I_minus_A_inv is not None:
            steady_state_term = I_minus_A_inv @ B[:,input_channel:input_channel+1]
            ss_gain = C @ steady_state_term + D[:,input_channel:input_channel+1]
            ss_gains[input_channel] = ss_gain
        else:
            # Estimate steady-state gain from the final values
            ss_gains[input_channel] = y[:,-1] - y[:,0]
    
    # Create plot if requested
    fig = None
    if plot:
        num_inputs = len(input_channels)
        num_outputs = C.shape[0]
        
        # Always create separate subplots for each input channel
        fig, axes = plt.subplots(num_inputs, 1, figsize=(12, 4 * num_inputs), squeeze=False)
        
        for i, input_channel in enumerate(input_channels):
            ax = axes[i, 0]
            
            for j in range(num_outputs):
                if output_names is not None and j < len(output_names):
                    label = output_names[j]
                else:
                    label = f'Output {j+1}'
                
                response = y_sim[input_channel][j, :] - y_sim[input_channel][j, 0]  # Remove initial offset
                ax.plot(np.arange(N), response, label=label)
                
                # Plot dotted line at steady-state value
                ss_value = ss_gains[input_channel][j, 0] if isinstance(ss_gains[input_channel], np.ndarray) else ss_gains[input_channel][j]
                ax.axhline(y=ss_value, color=f'C{j}', linestyle='--', alpha=0.7)
            
            ax.set_title(f'Step Response - Input Channel {input_channel+1}' if title is None else f'{title} - Input {input_channel+1}')
            ax.set_xlabel('Time Steps')
            ax.set_ylabel('Amplitude')
            ax.grid(True, alpha=0.3)
            ax.legend()
        plt.tight_layout()
    
    return


import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime

class IDDataTimeSeriesPlot:
    """Mixin with timeseries plotting functionality"""

    def plot_timeseries(self, figsize=(12, 8), time_unit='s', use_timestamps=True, 
                       title=None, grid=True, subplot_kwargs=None, show=True):
        """
        Plot the output and input signals of the system over time.
        
        Outputs are shown as line plots with markers at each sample point.
        Inputs are shown as step plots (zero-order hold) with markers at each sample point.
        
        Args:
            figsize (tuple, optional): Figure size as (width, height)
            time_unit (str, optional): Unit for time axis
            use_timestamps (bool, optional): Whether to use timestamps or sample indices for x-axis
            title (str, optional): Title for the figure
            grid (bool, optional): Whether to display grid lines
            subplot_kwargs (dict, optional): Additional arguments to pass to plt.subplots
            show (bool, optional): Whether to call plt.show() after creating the plot
        
        Returns:
            tuple: (fig, axs) - Figure and axes objects for further customization
        """
        if subplot_kwargs is None:
            subplot_kwargs = {}
        
        # Calculate the number of subplots needed
        n_plots = self.n_outputs + self.n_inputs
        
        # Create a figure with subplots
        fig, axs = plt.subplots(n_plots, 1, figsize=figsize, 
                               sharex=True, **subplot_kwargs)
        
        # Ensure axs is always a list/array even if n_plots=1
        axs = np.atleast_1d(axs)
        
        # Prepare x-axis data
        if use_timestamps and isinstance(self.timestamps[0], (datetime, np.datetime64)):
            # For datetime objects
            x = self.timestamps
            xlabel = 'Time'
        elif use_timestamps:
            # For numeric timestamps
            x = self.timestamps
            xlabel = f'Time [{time_unit}]'
        else:
            # Use sample indices
            x = np.arange(self.n_samples)
            xlabel = 'Sample'
        
        # Plot outputs as lines with markers
        for i in range(self.n_outputs):
            ax = axs[i]
            y_data = self.y[i, :]
            
            # Plot the output data as a continuous line with markers
            ax.plot(x, y_data, 'b-', linewidth=1.5)  # Line
            ax.plot(x, y_data, 'bo', markersize=3)   # Points
            
            # Label with name and unit if provided
            label = self.y_names[i]
            if self.y_units[i]:
                label += f" [{self.y_units[i]}]"
            
            ax.set_ylabel(label)
            
            if i == 0 and title:
                ax.set_title(title)
            elif i == 0:
                ax.set_title('System Outputs and Inputs')
            
            if grid:
                ax.grid(True, linestyle='--', alpha=0.7)
        
        # Plot inputs as steps with markers
        for i in range(self.n_inputs):
            ax = axs[i + self.n_outputs]
            u_data = self.u[i, :]
            
            # For step plot, we need to append an extra x value and reuse the last y value
            # to ensure the step extends to the end of the plot
            if len(x) > 1:
                # Calculate the step size to extend beyond the last point
                if isinstance(x[0], (datetime, np.datetime64)):
                    # For datetime x-axis, we need a different approach
                    # Use the average time delta between points to extend
                    time_deltas = np.diff(x)
                    if len(time_deltas) > 0:
                        avg_delta = np.mean(time_deltas)
                        x_extended = np.append(x, x[-1] + avg_delta)
                    else:
                        # Default extension if we can't calculate delta
                        x_extended = np.append(x, x[-1])
                else:
                    # For numeric x-axis
                    if len(x) > 1:
                        step = x[1] - x[0]
                        x_extended = np.append(x, x[-1] + step)
                    else:
                        x_extended = np.append(x, x[-1] + 1)
                
                u_extended = np.append(u_data, u_data[-1])
                
                # Plot the input data as steps
                ax.step(x_extended, u_extended, 'r-', linewidth=1.5, where='post')
                
                # Add points at each actual data point
                ax.plot(x, u_data, 'ro', markersize=3)
            else:
                # If only one point, just plot the point
                ax.plot(x, u_data, 'ro', markersize=5)
            
            # Label with name and unit if provided
            label = self.u_names[i]
            if self.u_units[i]:
                label += f" [{self.u_units[i]}]"
            
            ax.set_ylabel(label)
            
            if grid:
                ax.grid(True, linestyle='--', alpha=0.7)
        
        # Set x-label on the bottom subplot
        axs[-1].set_xlabel(xlabel)
        
        # Adjust layout
        plt.tight_layout()
        
        # Show the plot if requested
        if show:
            plt.show()
        
        return


import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
from matplotlib.backends.backend_pdf import PdfPages

from .residual_analysis_functions import (
    residual_statistics, plot_residual_statistics, normality_test,
    autocorrelation, cross_correlation, independence_test
)

def perform_residual_analysis(residuals, inputs=None, max_lag=20, alpha=0.05, 
                             save_report=False, report_path='residual_analysis_report.pdf',
                             show_plots=True):
    """
    Perform a comprehensive residual analysis for state space model validation.
    
    Parameters:
    -----------
    residuals : array-like
        The residuals from your state space model (measured - predicted).
    inputs : array-like or dict, optional
        The input variables used in your model. If numpy array, should be 2D where
        rows are different inputs and columns are timepoints.
    max_lag : int, optional
        Maximum lag to consider for correlation analyses.
    alpha : float, optional
        Significance level for statistical tests (default: 0.05).
    save_report : bool, optional
        Whether to save results to a PDF report.
    report_path : str, optional
        File path for the PDF report.
    show_plots : bool, optional
        Whether to display plots during analysis.
        
    Returns:
    --------
    dict
        Dictionary containing all analysis results and validation summary.
    """
    # Ensure residuals are numpy array
    residuals = np.asarray(residuals).flatten()
    n_samples = len(residuals)
    
    # Dictionary to store all results
    results = {'validation_summary': {}}
    
    # Set plot style
    plt.style.use('seaborn-v0_8-whitegrid')
    
    # Optional PDF report
    if save_report:
        pdf = PdfPages(report_path)
    
    # Create figure list to keep track
    figures = []
    
    # Print header for console output
    print("\n" + "="*80)
    print("COMPREHENSIVE RESIDUAL ANALYSIS FOR STATE SPACE MODEL VALIDATION")
    print("="*80)
    
    #######################################################
    # 1. Basic Statistical Analysis of Residuals
    #######################################################
    print("\n" + "="*80)
    print("1. STATISTICAL ANALYSIS OF RESIDUALS")
    print("="*80)

    # Calculate basic statistics
    stats_results = residual_statistics(residuals)
    results['statistics'] = stats_results
    
    print("\nBasic Statistics:")
    for key, value in stats_results.items():
        print(f"{key}: {value}")
    
    print("\nInterpretation:")
    # Check if mean is close to zero (within 2 standard errors)
    se = stats_results['std_dev'] / np.sqrt(n_samples)
    mean_ok = abs(stats_results['mean']) < 2 * se
    
    if mean_ok:
        print("✓ Mean is not significantly different from zero")
        print(f"  (mean: {stats_results['mean']:.6f}, 2*SE: {2*se:.6f})")
    else:
        print("✗ Mean is significantly different from zero")
        print(f"  (mean: {stats_results['mean']:.6f}, 2*SE: {2*se:.6f})")
        print("  This suggests model bias or systematic error")
    
    # Check skewness
    if abs(stats_results['skewness']) < 0.5:
        print("✓ Skewness is close to zero (symmetric distribution)")
    else:
        print(f"✗ Skewness is {stats_results['skewness']:.4f}, suggesting asymmetric residuals")
        print("  This may indicate nonlinear effects not captured by the model")
    
    # Check kurtosis (excess kurtosis close to 0)
    excess_kurtosis = stats_results['kurtosis'] - 3
    if abs(excess_kurtosis) < 1:
        print("✓ Kurtosis is close to normal distribution")
    else:
        if excess_kurtosis > 0:
            print(f"✗ Excess kurtosis is {excess_kurtosis:.4f}, indicating heavy tails")
            print("  This suggests presence of outliers or non-Gaussian behavior")
        else:
            print(f"✗ Excess kurtosis is {excess_kurtosis:.4f}, indicating light tails")
            print("  This suggests bounded or constrained behavior")
    
    # Visualize the statistical properties
    stat_fig = plot_residual_statistics(residuals)
    stat_fig.suptitle("Statistical Properties of Residuals", fontsize=16)
    
    if save_report:
        pdf.savefig(stat_fig)
    if show_plots:
        plt.show()
    else:
        plt.close(stat_fig)
    
    figures.append(stat_fig)
    
    #######################################################
    # 2. Normality Tests
    #######################################################
    print("\n" + "="*80)
    print("2. NORMALITY TESTS")
    print("="*80)
    
    # Perform normality tests
    norm_results = normality_test(residuals)
    results['normality'] = norm_results
    
    print("\nNormality Test Results:")
    for test_name, test_results in norm_results.items():
        print(f"\n{test_name}:")
        for key, value in test_results.items():
            print(f"  {key}: {value}")
    
    # Get p-value from Shapiro-Wilk test for summary
    shapiro_p = norm_results['shapiro_wilk']['p_value']
    normality_ok = shapiro_p > alpha
    
    print("\nInterpretation:")
    if normality_ok:
        print(f"✓ Shapiro-Wilk test (p={shapiro_p:.4f} > {alpha}) suggests residuals are normally distributed")
        print("  This is consistent with white noise assumptions")
    else:
        print(f"✗ Shapiro-Wilk test (p={shapiro_p:.4f} < {alpha}) suggests residuals are NOT normally distributed")
        print("  This may indicate nonlinear behavior, outliers, or heteroscedasticity")
    
    #######################################################
    # 3. Autocorrelation Analysis
    #######################################################
    print("\n" + "="*80)
    print("3. AUTOCORRELATION ANALYSIS")
    print("="*80)
    
    # Perform autocorrelation analysis
    acf_fig, acf_values, lb_results = autocorrelation(residuals, max_lag=max_lag, alpha=alpha)
    acf_fig.suptitle("Autocorrelation Analysis of Residuals", fontsize=16)
    results['autocorrelation'] = {
        'acf_values': acf_values,
        'ljung_box': lb_results
    }
    
    if save_report:
        pdf.savefig(acf_fig)
    if show_plots:
        plt.show()
    else:
        plt.close(acf_fig)
    
    figures.append(acf_fig)
    
    # Check Ljung-Box test results at specific lags
    lag_indices = [4, 9, 14]  # Lags 5, 10, 15
    lb_significant = False
    
    print("\nLjung-Box Test Results (selected lags):")
    for idx in lag_indices:
        if idx < len(lb_results):
            lag = lb_results['lag'][idx]
            p_val = lb_results['lb_pvalue'][idx]
            significant = p_val < alpha
            if significant:
                lb_significant = True
            print(f"Lag {lag}: p-value = {p_val:.6f} ({'Significant' if significant else 'Not significant'})")
    
    print("\nInterpretation:")
    autocorr_ok = not lb_significant
    if autocorr_ok:
        print("✓ No significant autocorrelation detected by Ljung-Box test")
        print("  Residuals appear to be white noise (good)")
    else:
        print("✗ Significant autocorrelation detected by Ljung-Box test")
        print("  This suggests your model doesn't capture all dynamics in the data")
        print("  Consider increasing model order or adding relevant states")
    
    #######################################################
    # 4. Cross-correlation Analysis (if inputs provided)
    #######################################################
    crosscorr_ok = True
    if inputs is not None:
        print("\n" + "="*80)
        print("4. CROSS-CORRELATION ANALYSIS")
        print("="*80)
        
        # Perform cross-correlation analysis
        ccf_fig, ccf_values = cross_correlation(residuals, inputs, max_lag=max_lag, alpha=alpha)
        ccf_fig.suptitle("Cross-correlation: Residuals vs Inputs", fontsize=16)
        results['cross_correlation'] = ccf_values
        
        if save_report:
            pdf.savefig(ccf_fig)
        if show_plots:
            plt.show()
        else:
            plt.close(ccf_fig)
        
        figures.append(ccf_fig)
        
        # Check if any significant cross-correlations
        print("\nSignificant Cross-correlations:")
        critical_value = stats.norm.ppf(1 - alpha/2) / np.sqrt(n_samples)
        for name in ccf_values:
            max_ccf = np.max(np.abs(ccf_values[name]))
            significant = max_ccf > critical_value
            if significant:
                crosscorr_ok = False
                max_idx = np.argmax(np.abs(ccf_values[name]))
                lag = max_idx - max_lag if max_idx - max_lag < 0 else max_idx - max_lag
                print(f"✗ {name}: Significant at lag {lag} (value: {ccf_values[name][max_idx]:.4f})")
        
        if crosscorr_ok:
            print("✓ No significant cross-correlation between residuals and inputs")
            print("  Model appears to capture input-output relationships adequately")
        else:
            print("\nInterpretation:")
            print("✗ Significant cross-correlation detected")
            print("  This suggests the model doesn't fully capture how inputs affect outputs")
            print("  Consider adjusting system structure, adding states, or modifying input integration")
    
    #######################################################
    # 5. Independence Tests
    #######################################################
    print("\n" + "="*80)
    print("5. INDEPENDENCE TESTS")
    print("="*80)
    
    # Perform independence tests
    indep_results = independence_test(residuals, max_lag=max_lag)
    results['independence'] = indep_results
    
    # Check Ljung-Box test at lag 10
    lb_lag10_idx = min(9, len(indep_results['ljung_box_test']['lags'])-1)
    lb_lag10_pval = indep_results['ljung_box_test']['p_values'][lb_lag10_idx]
    lb_indep_ok = lb_lag10_pval > alpha
    
    # Check McLeod-Li test at lag 10 (for ARCH effects)
    ml_lag10_idx = min(9, len(indep_results['mcleod_li_test']['lags'])-1)
    ml_lag10_pval = indep_results['mcleod_li_test']['p_values'][ml_lag10_idx]
    ml_indep_ok = ml_lag10_pval > alpha
    
    # Runs test
    runs_pval = indep_results['runs_test']['p_value']
    # Handle potential NaN values
    if np.isnan(runs_pval):
        runs_ok = True  # Consider it passed if we couldn't compute it
        print("Note: Runs test could not be computed reliably with this data")
    else:
        runs_ok = runs_pval > alpha
    
    print("\nLjung-Box Test (lag 10):")
    print(f"p-value: {lb_lag10_pval:.6f} ({'Not significant' if lb_indep_ok else 'Significant'})")
    
    print("\nMcLeod-Li Test (lag 10):")
    print(f"p-value: {ml_lag10_pval:.6f} ({'Not significant' if ml_indep_ok else 'Significant'})")
    
    print("\nRuns Test:")
    print(f"p-value: {runs_pval:.6f} ({'Not significant' if runs_ok else 'Significant'})")
    
    print("\nInterpretation:")
    independence_ok = lb_indep_ok and ml_indep_ok and runs_ok
    
    if independence_ok:
        print("✓ All independence tests passed")
        print("  Residuals appear to be random and independent")
    else:
        print("✗ One or more independence tests failed")
        if not lb_indep_ok:
            print("  - Ljung-Box test suggests linear dependencies in residuals")
        if not ml_indep_ok:
            print("  - McLeod-Li test suggests nonlinear dependencies (ARCH effects)")
        if not runs_ok:
            print("  - Runs test suggests non-random sequence of residuals")
    
    #######################################################
    # 6. Overall Model Validation Decision
    #######################################################
    print("\n" + "="*80)
    print("6. MODEL VALIDATION SUMMARY")
    print("="*80)
    
    # Store validation results
    validation = {
        'mean_ok': mean_ok,
        'normality_ok': normality_ok,
        'autocorr_ok': autocorr_ok,
        'crosscorr_ok': crosscorr_ok,
        'independence_ok': independence_ok
    }
    results['validation_summary'] = validation
    
    print("\nValidation Checks:")
    print(f"Mean close to zero: {'✓' if mean_ok else '✗'}")
    print(f"Residuals normally distributed: {'✓' if normality_ok else '✗'}")
    print(f"No significant autocorrelation: {'✓' if autocorr_ok else '✗'}")
    print(f"No significant cross-correlation with inputs: {'✓' if crosscorr_ok else '✗'}")
    print(f"Residuals pass independence tests: {'✓' if independence_ok else '✗'}")
    
    # Overall decision
    model_ok = mean_ok and normality_ok and autocorr_ok and crosscorr_ok and independence_ok
    
    print("\nOverall Model Validation:", end=" ")
    if model_ok:
        print("✓ PASSED - Model appears to be adequate")
    else:
        print("✗ FAILED - Model may need improvement")
        print("\nRecommended actions:")
        if not mean_ok:
            print("- Check for bias in the model structure or measurement process")
        if not normality_ok:
            print("- Investigate outliers, consider nonlinear transformations")
            print("- Check for heteroscedasticity or time-varying behavior")
        if not autocorr_ok:
            print("- Increase model order or add relevant state variables")
            print("- Consider adding time-delay terms or memory effects")
        if not crosscorr_ok:
            print("- Improve how the model captures input-output relationships")
            print("- Check for nonlinearities in input effects")
        if not independence_ok:
            print("- Look for missing inputs or disturbances")
            print("- Consider nonlinear model structures")
    
    # Generate a PDF report if requested
    if save_report:
        # Create a summary page
        fig_summary = plt.figure(figsize=(8.5, 11))
        plt.axis('off')
        
        title = "State Space Model Residual Analysis Summary"
        
        summary_text = [
            f"Number of samples: {n_samples}",
            f"Residual mean: {stats_results['mean']:.6f}",
            f"Residual std dev: {stats_results['std_dev']:.6f}",
            f"Mean test: {'Passed' if mean_ok else 'Failed'}",
            f"Normality test: {'Passed' if normality_ok else 'Failed'}",
            f"Autocorrelation test: {'Passed' if autocorr_ok else 'Failed'}",
            f"Cross-correlation test: {'Passed' if crosscorr_ok else 'Failed'}",
            f"Independence test: {'Passed' if independence_ok else 'Failed'}",
            f"\nOverall validation: {'Passed' if model_ok else 'Failed'}"
        ]
        
        plt.text(0.5, 0.95, title, fontsize=16, ha='center', va='top', transform=fig_summary.transFigure)
        plt.text(0.1, 0.85, "\n".join(summary_text), fontsize=12, ha='left', va='top', transform=fig_summary.transFigure)
        
        pdf.savefig(fig_summary)
        plt.close(fig_summary)
        
        # Close the PDF
        pdf.close()
        print(f"\nDetailed analysis report saved to {report_path}")
    
    return


# Example usage
if __name__ == "__main__":
    # Generate synthetic data for demonstration
    np.random.seed(42)
    n_samples = 500
    
    # Well-behaved residuals (white noise)
    good_residuals = np.random.normal(0, 1, n_samples)
    
    # Problematic residuals (autocorrelated)
    bad_residuals = np.zeros(n_samples)
    for i in range(1, n_samples):
        bad_residuals[i] = 0.7 * bad_residuals[i-1] + np.random.normal(0, 0.7)
    
    # Generate inputs
    n_inputs = 2
    inputs = np.zeros((n_inputs, n_samples))
    inputs[0, :] = np.random.normal(0, 1, n_samples)  # Random noise
    inputs[1, :] = np.sin(np.linspace(0, 10*np.pi, n_samples))  # Sine wave
    
    # Example 1: Good model
    print("\n\nANALYSIS OF GOOD MODEL RESIDUALS")
    results_good = perform_residual_analysis(
        good_residuals, 
        inputs=inputs,
        save_report=True, 
        report_path="good_model_residual_analysis.pdf",
        show_plots=True
    )
    
    # Example 2: Model with issues
    print("\n\nANALYSIS OF PROBLEMATIC MODEL RESIDUALS")
    results_bad = perform_residual_analysis(
        bad_residuals, 
        inputs=inputs,
        save_report=True, 
        report_path="bad_model_residual_analysis.pdf",
        show_plots=True
    )


import numpy as np
import warnings
import copy

from buildingsysid.calculate.solvers.least_squares_solver import LeastSquaresSolver
from buildingsysid.criterion_of_fit.objective_functions import StandardObjective

class OptimizationManager:
    """
    Coordinator class for system identification optimization problems
    with enhanced error handling and robustness.
    
    This class coordinates the optimization process by:
    - Connecting the model with the objective function and data
    - Setting up the bounds for optimization
    - Preparing the problem for external solvers
    - Handling numerical issues during optimization
    """
    
    def __init__(self, model_structure, data, objective=None, solver=None):
        """
        Initialize the optimization problem.
        
        Args:
            model_structure: Instance of model structure
            data: IdData object with the data to use for identification
            objective: ObjectiveFunction instance (if None, it must be set later)
            solver: Solver instance (if None, a default LeastSquaresSolver will be created)
        """
        
        # Create a working copy of the model structure
        self.model = copy.deepcopy(model_structure)
        
        # Set the iddata
        self.data = data
        
        # Set the objective
        self.objective = objective if objective is not None else StandardObjective()
        
        # Set the solver
        self.solver = solver if solver is not None else LeastSquaresSolver()
        
        # Configure feedback parameters if needed
        if hasattr(self.objective, 'kstep'):
            # If objective uses feedback (kstep > 0), enable feedback parameters
            self.model.set_feedback_mode(use_feedback=self.objective.kstep > 0)
        
        # Results storage
        self.result = None
        self.attempts = 0  # Track the number of attempts made
        self.max_attempts = 5  # Maximum number of attempts to try
    
    
    # =================================================================
    # Method to Wrap Around the Objective (To Fit the Solver)
    # =================================================================    
    def objective_wrapper(self, par):
        """
        Wrapper around the objective function that handles numerical issues.
        
        Args:
            par: Parameter vector
            
        Returns:
            Array of residuals from the objective function
        """
        if self.objective is None:
            raise ValueError("No objective function specified")
        
        try:
            # Try to evaluate the objective function
            residuals = self.objective(self.model, self.data, par)
            
            # Check for NaN or Inf values
            if np.any(np.isnan(residuals)) or np.any(np.isinf(residuals)):
                # Return a high penalty value instead of NaN/Inf
                return np.ones_like(residuals) * 1e10
            
            return residuals
        except Exception as e:
            # If there's an error, return a high penalty value
            print(f"Warning: Error in objective function: {e}")
            # Return a large residual vector
            return np.ones(self.data.y.size) * 1e10
    
    
    # =================================================================
    # Method to Calculate RMSE from Residuals
    # =================================================================
    def calculate_rmse(self, residuals):
        """
        Calculate RMSE directly from residuals.
        
        Args:
            residuals: Vector of residuals
            
        Returns:
            RMSE value
        """
        return np.sqrt(np.mean(residuals**2))
    
    
    # =================================================================
    # Method to Generate Initial Guesses
    # =================================================================    
    def generate_initial_parameters(self, strategy='random'):
        """
        Generate initial parameters using various strategies.
        
        Args:
            strategy: Strategy to use ('random', 'zeros', 'ones', 'middle', 'perturbed', 'biased_random')
                    or a numpy array/list of specific initial values
            
        Returns:
            Array of initial parameter values
        """
        # Check if strategy is already a parameter vector
        if isinstance(strategy, (np.ndarray, list)):
            # Convert to numpy array if needed
            return np.array(strategy)
        
        # Get bounds
        lower_bounds, upper_bounds = self.model.get_parameter_bounds()
        
        # Number of free parameters
        n_params = self.model.n_free_parameters
        
        if strategy == 'random':
            # Random initialization (uniform distribution between sensible bounds)
            lower_init = np.where(np.isfinite(lower_bounds), lower_bounds, -1.0)
            upper_init = np.where(np.isfinite(upper_bounds), upper_bounds, 1.0)
            
            # Add small margin to avoid boundary values
            margin = 0.05 * (upper_init - lower_init)
            lower_init = lower_init + margin
            upper_init = upper_init - margin
            
            result = np.random.uniform(low=lower_init, high=upper_init, size=n_params)
        
        elif strategy == 'zeros':
            result = np.zeros(n_params)
            
            # Check if zeros are within bounds and adjust if needed
            for i in range(n_params):
                if result[i] < lower_bounds[i]:
                    result[i] = lower_bounds[i]
                elif result[i] > upper_bounds[i]:
                    result[i] = upper_bounds[i]
        
        elif strategy == 'ones':
            result = np.ones(n_params)
            
            # Check if ones are within bounds and adjust if needed
            for i in range(n_params):
                if result[i] < lower_bounds[i]:
                    result[i] = lower_bounds[i]
                elif result[i] > upper_bounds[i]:
                    result[i] = upper_bounds[i]
        
        elif strategy == 'middle':
            # Middle of bounds range
            lower_init = np.where(np.isfinite(lower_bounds), lower_bounds, -1.0)
            upper_init = np.where(np.isfinite(upper_bounds), upper_bounds, 1.0)
            result = 0.5 * (lower_init + upper_init)
        
        elif strategy == 'perturbed':
            # Start with zeros but add small perturbations
            result = np.random.normal(0, 0.1, size=n_params)
            
            # Check bounds
            for i in range(n_params):
                if result[i] < lower_bounds[i]:
                    result[i] = lower_bounds[i]
                elif result[i] > upper_bounds[i]:
                    result[i] = upper_bounds[i]
        
        elif strategy == 'black_box':
            # Random values biased toward typical values but respecting bounds
            result = np.zeros(n_params)
            
            for i in range(n_params):
                idx = self.model.free_indices[i]
                param_name = self.model.param_dict[idx][0]
                
                # Get bounds for this parameter
                lower = lower_bounds[i]
                upper = upper_bounds[i]
                
                # Ensure bounds are finite for sampling
                if not np.isfinite(lower):
                    lower = -1.0
                if not np.isfinite(upper):
                    upper = 1.0
                    
                # Bias initial guesses based on parameter type
                if param_name.startswith('a'):  # State matrix elements
                    # Dynamic parameters often negative for stable systems
                    proposed = np.random.uniform(-1.0, -0.1)
                elif param_name.startswith('b'):  # Input matrix elements
                    # Input gains often small positive
                    proposed = np.random.uniform(0.1, 1.0)
                elif param_name.startswith('k'):  # Feedback matrix elements
                    # Feedback parameters are often small
                    proposed = np.random.uniform(-0.2, 0.2)
                else:
                    # Default random
                    proposed = np.random.uniform(-1.0, 1.0)
                
                # Enforce bounds
                result[i] = np.clip(proposed, lower, upper)
                
        elif strategy == 'grey_box':
            result = np.zeros(n_params)
            
            for i in range(n_params):
                idx = self.model.free_indices[i]
                param_name = self.model.param_dict[idx][0]
                
                # Get bounds for this parameter
                lower = lower_bounds[i]
                upper = upper_bounds[i]
                
                # Ensure bounds are finite for sampling
                if not np.isfinite(lower):
                    lower = -1.0
                if not np.isfinite(upper):
                    upper = 1.0
                    
                # Physics-informed initial values
                if param_name.startswith('H'):  # specific heat transfer
                    proposed = np.random.uniform(0.1, 1.0)
                elif param_name.startswith('C'):  # specific heat capacity
                    proposed = np.random.uniform(0.1, 1.0)
                elif param_name.startswith('ws'):  # solar input
                    proposed = np.random.uniform(0.1, 1.0)
                elif param_name.startswith('wh'):  # heat input
                    proposed = np.random.uniform(0.1, 1.0)
                elif param_name.startswith('k'):  # Feedback matrix elements
                    proposed = np.random.uniform(-0.2, 0.2)
                elif param_name.startswith('x'):  # Initial state (temperatures)
                    proposed = np.random.uniform(18, 22)
                else:
                    # Default random
                    proposed = np.random.uniform(-1.0, 1.0)
                
                # Enforce bounds
                result[i] = np.clip(proposed, lower, upper)
        
        else:
            raise ValueError(f"Unknown initialization strategy: {strategy}")
            
        return result
    
    
    # =================================================================
    # Method to Solve Optimization Problem
    # =================================================================     
    def solve(self, x0=None, max_attempts=5, initialization_strategies=["random"], max_rmse=None):
        """
        Solve the optimization problem using the provided solver with robustness.
        
        Args:
            x0: Initial parameter values (if None, generated automatically)
            max_attempts: Maximum number of attempts per strategy if the strategy fails
            initialization_strategies: List of strategies to try.
                Can include: 'random', 'zeros', 'middle', 'perturbed', 'ones', 'biased_random'
                or numpy arrays of specific initial values.
                Each strategy will be tried until success or max_attempts is reached.
                All strategies will be tried regardless of max_attempts.
            max_rmse: Maximum acceptable RMSE (if None, continues all attempts)
                Solutions with RMSE below this threshold are considered satisfactory.
            
        Returns:
            A new, updated model structure instance (the original model is not modified)
        """
        
        # Get bounds
        lower_bounds, upper_bounds = self.model.get_parameter_bounds()
        
        # Set maximum attempts per strategy
        self.max_attempts = max_attempts
        
        strategies = initialization_strategies
        
        # If x0 is provided, add it as the first strategy
        if x0 is not None:
            strategies = [x0] + list(strategies)
        
        # Try optimization with different initial parameters
        best_result = None
        lowest_cost = float('inf')
        lowest_rmse = float('inf')
        self.attempts = 0
        
        # Iterate through all strategies
        for strategy_idx, strategy in enumerate(strategies):
            strategy_success = False
            strategy_attempts = 0
            
            # Determine the strategy name for display purposes
            if isinstance(strategy, (np.ndarray, list)):
                strategy_name = "custom array"
            else:
                strategy_name = strategy
            
            print(f"\nTrying strategy {strategy_idx+1}/{len(strategies)}: {strategy_name}")
            
            # Try this strategy up to max_attempts times
            while not strategy_success and strategy_attempts < max_attempts:
                strategy_attempts += 1
                self.attempts += 1
                
                # Generate initial parameters for this attempt
                x0_attempt = self.generate_initial_parameters(strategy)
                
                print(f"  Attempt {strategy_attempts}/{max_attempts} with {strategy_name} initialization")
                
                try:
                    # Catch warnings during optimization
                    with warnings.catch_warnings(record=True) as w:
                        warnings.simplefilter("always")
                        
                        # Call the solver with the problem components
                        result = self.solver.solve(
                            objective_fn=self.objective_wrapper,
                            x0=x0_attempt, 
                            bounds=(lower_bounds, upper_bounds)
                        )
                    
                    # Check if optimization succeeded and improved the objective
                    success = hasattr(result, 'success') and result.success
                    cost = result.cost
                    
                    # Calculate RMSE for this result directly from residuals
                    if hasattr(result, 'fun'):
                        current_rmse = self.calculate_rmse(result.fun)
                    else:
                        # Fallback in case residuals aren't available
                        current_rmse = float('inf')
                    
                    # Add RMSE to the result
                    result.rmse = current_rmse
                    
                    print(f"    Success: {success}, RMSE: {current_rmse:.6f}, Cost (SSE): {cost:.6e}")
                    
                    # Mark this strategy as successful
                    if success:
                        strategy_success = True
                        
                        # Update best result if this attempt is better
                        if cost < lowest_cost:
                            lowest_cost = cost
                            lowest_rmse = current_rmse
                            best_result = result
                            print(f"    → New best result (RMSE: {lowest_rmse:.6f})")
                            
                            # Break early if we have a good solution
                            if max_rmse is not None and current_rmse < max_rmse:
                                print(f"    → Found satisfactory solution (RMSE < {max_rmse:.6f})")
                                break
                    
                except Exception as e:
                    error_str = str(e)
                    print(f"    Optimization attempt failed: {error_str}")
                    
                    # If the error is about NaNs or Infs, show that we'll retry
                    if "infs or NaNs" in error_str:
                        print(f"    → Detected NaN/Inf error. Will retry with a new {strategy_name} initialization.")
                        # Continue the while loop to retry this strategy
        
        # Create an optimized model as a copy
        optimized_model = None
        
        # Use the best result
        if best_result is not None:
            self.result = best_result
            
            # Create a new model instance with the optimized parameters
            optimized_model = copy.deepcopy(self.model)
            
            if hasattr(best_result, 'x'):
                # Store the result in the model (but don't modify the original)
                optimized_model.result = best_result                
                print("\nOptimization completed successfully")
                print(f"Final RMSE: {lowest_rmse:.6f}")
                print(f"Final SSE: {lowest_cost:.6e}")
        else:
            print(f"\nWarning: All {self.attempts} optimization attempts across all strategies failed.")
            
            # Create a dummy result
            class FailedResult:
                def __init__(self):
                    self.success = False
                    self.message = "All optimization attempts failed"
                    self.x = None
                    self.fun = np.array([float('inf')])  # Placeholder for residuals
                    self.rmse = float('inf')
            
            self.result = FailedResult()
            # Return a copy of the original model in case of failure
            optimized_model = copy.deepcopy(self.model)
        
        return optimized_model


import numpy as np
import os

class StateSpace:
    
    def __init__(self, A, B, C, D, K=None, x0=None, samplingTime=0):
        self.A = A
        self.B = B
        self.C = C
        self.D = D
        self.K = K
        self.x0 = x0
    
        self.samplingTime = samplingTime  # 0 = continuous time, otherwise discrete
    
    
    
    def save(self, filename):
        """
        Save the state-space model to a file using NumPy's .npz format.
        
        Parameters:
        -----------
        filename : str
            The filename to save to. If no extension is provided, .npz will be added.
        """
        # Add .npz extension if not already present
        if not filename.endswith('.npz'):
            filename = filename + '.npz'
        
        # Convert all attributes to NumPy arrays or basic Python types
        save_dict = {
            'A': self.A,
            'B': self.B,
            'C': self.C,
            'D': self.D,
            'samplingTime': self.samplingTime
        }
        
        # Save K and x0 only if they exist
        if self.K is not None:
            save_dict['K'] = self.K
        
        if self.x0 is not None:
            save_dict['x0'] = self.x0
        
        # Save to file
        np.savez(filename, **save_dict)
        print(f"Model saved to {filename}")
    
    @classmethod
    def load(cls, filename):
        """
        Load a state-space model from a .npz file.
        
        Parameters:
        -----------
        filename : str
            The filename to load from.
            
        Returns:
        --------
        StateSpace
            The loaded state-space model.
        """
        # Add .npz extension if not already present
        if not filename.endswith('.npz'):
            filename = filename + '.npz'
        
        # Load the data
        data = np.load(filename, allow_pickle=True)
        
        # Extract model parameters
        A = data['A']
        B = data['B'] 
        C = data['C']
        D = data['D']
        samplingTime = float(data['samplingTime'])
        
        # Handle optional parameters
        K = data['K'] if 'K' in data else None
        x0 = data['x0'] if 'x0' in data else None
        
        # Create and return the model
        model = cls(A, B, C, D, K, x0, samplingTime)
        print(f"Model loaded from {filename}")
        return model


# Example usage:
if __name__ == "__main__":
    # Define a simple model
    A = np.array([[0.8, 0.1], [0, 0.9]])
    B = np.array([[0.5], [1.0]])
    C = np.array([[1.0, 0.2]])
    D = np.array([[0]])
    K = np.array([[0.1, 0.2]])
    
    # Create model
    model = StateSpace(A, B, C, D, K=K, samplingTime=0.1)
    
    # Save model
    model.save("my_model")
    
    # Load model
    loaded_model = StateSpace.load("my_model")
    
    # Verify it loaded correctly
    print("\nOriginal model:")
    print(f"A: {model.A}")
    print(f"K: {model.K}")
    print(f"samplingTime: {model.samplingTime}")
    
    print("\nLoaded model:")
    print(f"A: {loaded_model.A}")
    print(f"K: {loaded_model.K}")
    print(f"samplingTime: {loaded_model.samplingTime}")